@@ -1,9 +1,17 @@
 use std::fs::File;
 use std::io::BufReader;
 use std::io::Write as _;
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 
 fn get_git_version() -> Result<String, Box<dyn std::error::Error>> {
+    if Path::new(".git/HEAD").exists() {
+        println!("cargo:rerun-if-changed=.git/HEAD");
+        let r = std::fs::read_to_string(".git/HEAD").unwrap();
+        let mut r = r.split_ascii_whitespace();
+        r.next();
+        let branch = r.next().unwrap();
+        println!("cargo:rerun-if-changed=.git/{}", branch);
+    }
     let git_version = std::process::Command::new("git").arg("describe").output()?;
     let mut git_version = String::from_utf8(git_version.stdout)?;
     git_version.pop();
@@ -14,21 +22,24 @@ fn main() {
     let full_version =
         get_git_version().unwrap_or_else(|_| std::env::var("CARGO_PKG_VERSION").unwrap());
 
-    // commenting this out as it would make the life hard for linters
-    //assert!(
-    //    full_version.starts_with(&std::env::var("CARGO_PKG_VERSION").unwrap()),
-    //    "latest git tag does not match the version set in cargo"
-    //);
+    assert!(
+        full_version.starts_with(&std::env::var("CARGO_PKG_VERSION").unwrap()),
+        "latest git tag does not match the version set in cargo: {} vs {}",
+        full_version,
+        std::env::var("CARGO_PKG_VERSION").unwrap()
+    );
 
     println!(
         "cargo:rustc-env=XASKPASS_BUILD_FULL_VERSION={}",
         full_version
     );
 
-    // Because clippy is so slow otherwise
-    let out_path = match std::fs::canonicalize("pregen") {
-        Err(_) => PathBuf::from(std::env::var("OUT_DIR").unwrap()),
-        Ok(path) => path,
+    let out_path = match std::env::var_os("XASKPASS_BUILDDIR") {
+        Some(path) => std::fs::canonicalize(path).unwrap(),
+        None => match std::fs::canonicalize("pregen") {
+            Err(_) => PathBuf::from(std::env::var_os("OUT_DIR").unwrap()),
+            Ok(path) => path,
+        },
     };
 
     let mut man = std::fs::read_to_string("xaskpass.man.in").unwrap();
