@@ -129,6 +129,7 @@ impl std::str::FromStr for Rgba {
 pub struct Config {
     pub title: String,
     pub grab_keyboard: bool,
+    pub show_hostname: bool,
     pub resizable: bool,
     pub depth: u8,
     pub dialog: Dialog,
@@ -139,8 +140,9 @@ impl Default for Config {
         Self {
             title: NAME.into(),
             grab_keyboard: false,
+            show_hostname: false,
             resizable: false,
-            depth: 24,
+            depth: 32,
             dialog: Dialog::default(),
         }
     }
@@ -195,7 +197,7 @@ impl Default for Dialog {
         Self {
             foreground: "#5c616c".parse().unwrap(),
             indicator_label_foreground: "#5c616c".parse().unwrap(),
-            background: "#f5f6f7".parse().unwrap(),
+            background: "#f5f6f7ee".parse().unwrap(),
             label: "Please enter your authentication passphrase:".into(),
             indicator_label: "Secret:".into(),
             input_timeout: Some(30),
