@@ -4,12 +4,11 @@ use std::ops::{Deref, DerefMut};
 use std::pin::Pin;
 use std::time::Duration;
 
-use log::{debug, trace};
+use log::{debug, log_enabled, trace};
 use rand::seq::SliceRandom as _;
 use tokio::time::{sleep, Instant, Sleep};
 
 use super::Pattern;
-use crate::backbuffer::FrameId;
 use crate::config;
 use crate::errors::Result;
 use crate::secret::{Passphrase, SecBuf};
@@ -147,8 +146,11 @@ impl Base {
             self.width + 2.0,
             self.height + 2.0,
         );
+        cr.save().unwrap();
+        cr.set_operator(cairo::Operator::Source);
         cr.set_source(background).unwrap();
         cr.fill().unwrap();
+        cr.restore().unwrap();
     }
 
     fn blink(
@@ -176,6 +178,7 @@ impl Base {
             cr.stroke().unwrap();
         } else {
             cr.rectangle(x - 1.0, y - 1.0, 3.0, height + 2.0);
+            cr.set_operator(cairo::Operator::Source);
             cr.set_source(bg.unwrap_or(&self.background)).unwrap();
             cr.fill().unwrap();
         };
@@ -256,8 +259,9 @@ pub struct Circle {
     animation_distance: f64,
     rotation: f64,
     lock_color: Pattern,
-    last_animation_serial: Option<FrameId>,
     oldlen: usize,
+    old_timestamp: Option<Instant>,
+    paint_pending: bool,
 }
 
 impl Deref for Circle {
@@ -309,11 +313,12 @@ impl Circle {
             frame_increment: frame_increment_start,
             frame_increment_start,
             frame_increment_gain: circle.rotation_speed_gain,
-            last_animation_serial: None,
             angle: 2.0 * std::f64::consts::PI / indicator_count as f64,
             animation_distance: 0.0,
             rotation: 0.0,
             oldlen: 0,
+            old_timestamp: None,
+            paint_pending: false,
         }
     }
 
@@ -344,62 +349,68 @@ impl Circle {
 
     fn init_rotation(&mut self) {
         trace!("run animation");
-        let full_round = 2.0 * std::f64::consts::PI;
-        self.rotation %= full_round;
+        const FULL_ROUND: f64 = 2.0 * std::f64::consts::PI;
+        self.rotation %= FULL_ROUND;
         self.animation_distance += (self.pass.len as i64 - self.oldlen as i64) as f64
             * (self.angle / self.indicator_count as f64);
         self.oldlen = self.pass.len;
-        if self.animation_distance > 2.0 * full_round {
-            self.animation_distance = self.animation_distance % full_round + full_round;
-        } else if self.animation_distance < 2.0 * -full_round {
-            self.animation_distance = self.animation_distance % full_round - full_round;
+        if self.animation_distance.abs() > 2.0 * FULL_ROUND {
+            self.animation_distance %= FULL_ROUND;
+            if self.animation_distance > 0.0 {
+                self.animation_distance += FULL_ROUND;
+            } else {
+                self.animation_distance -= FULL_ROUND;
+            }
+        }
+        if !self.paint_pending && self.animation_distance != 0.0 {
+            self.animate_frame();
         }
     }
 
-    pub fn on_displayed(&mut self, serial: FrameId) -> bool {
-        trace!(
-            "serial {:?}, last_animation_serial {:?}",
-            serial,
-            self.last_animation_serial
-        );
-        if let Some(s) = self.last_animation_serial {
-            assert!(self.animation_distance != 0.0);
-            if serial == s {
-                self.last_animation_serial = None;
-                let mut animation_running = true;
-                if self.animation_distance > 0.0 {
-                    self.rotation += self.frame_increment.min(self.animation_distance);
-                    self.animation_distance -= self.frame_increment;
-                    if self.animation_distance <= 0.0 {
-                        animation_running = false;
-                    }
-                    trace!(
-                        "animation_distance {}, rotation {}",
-                        self.animation_distance,
-                        self.rotation
-                    );
-                } else {
-                    self.rotation -= self.frame_increment.min(-self.animation_distance);
-                    self.animation_distance += self.frame_increment;
-                    if self.animation_distance >= 0.0 {
-                        animation_running = false;
-                    }
-                }
+    pub fn set_next_frame(&mut self) {
+        trace!("set_next_frame");
+        assert!(!self.dirty && !self.dirty_blink);
+        if self.animation_distance == 0.0 {
+            trace!("not animating");
+            return;
+        }
+        self.animate_frame();
+    }
 
-                if animation_running {
-                    self.frame_increment *= self.frame_increment_gain;
-                } else {
-                    self.frame_increment = self.frame_increment_start;
-                    self.animation_distance = 0.0;
-                }
-
-                self.dirty = true;
-                return true;
-            } else {
-                trace!("last animated frame might not have been shown yet");
+    fn animate_frame(&mut self) {
+        assert!(!self.paint_pending);
+        self.paint_pending = true;
+        let mut animation_running = true;
+        if self.animation_distance > 0.0 {
+            self.rotation += self.frame_increment.min(self.animation_distance);
+            self.animation_distance -= self.frame_increment;
+            if self.animation_distance <= 0.0 {
+                animation_running = false;
+            }
+            trace!(
+                "animation_distance {}, rotation {}",
+                self.animation_distance,
+                self.rotation
+            );
+        } else {
+            self.rotation -= self.frame_increment.min(-self.animation_distance);
+            self.animation_distance += self.frame_increment;
+            if self.animation_distance >= 0.0 {
+                animation_running = false;
             }
         }
-        false
+
+        if animation_running {
+            self.frame_increment *= self.frame_increment_gain;
+        } else {
+            self.frame_increment = self.frame_increment_start;
+            self.animation_distance = 0.0;
+        }
+
+        self.dirty = true;
+        if log_enabled!(log::Level::Debug) {
+            self.old_timestamp = Some(Instant::now());
+        }
     }
 
     fn blink(&self, cr: &cairo::Context) {
@@ -414,15 +425,9 @@ impl Circle {
         );
     }
 
-    pub fn set_painted(&mut self, serial: FrameId) {
-        trace!(
-            "set_painted serial {:?}, animation_distance {}",
-            serial,
-            self.animation_distance
-        );
-        if self.animation_distance != 0.0 && self.last_animation_serial.is_none() {
-            self.last_animation_serial = Some(serial);
-        }
+    pub fn set_painted(&mut self) {
+        trace!("set_painted paint_pending {:?}", self.paint_pending,);
+        self.paint_pending = false;
         self.base.set_painted()
     }
 
