@@ -13,7 +13,6 @@ use tokio::time::{sleep, Instant, Sleep};
 use x11rb::protocol::xproto;
 use zeroize::Zeroize;
 
-use crate::backbuffer::FrameId;
 use crate::bail;
 use crate::config;
 use crate::config::{IndicatorType, Rgba};
@@ -275,19 +274,19 @@ impl Indicator {
         }
     }
 
-    pub fn set_painted(&mut self, serial: FrameId) {
+    pub fn set_painted(&mut self) {
         match self {
             Self::Strings(i) => i.set_painted(),
-            Self::Circle(i) => i.set_painted(serial),
+            Self::Circle(i) => i.set_painted(),
             Self::Classic(i) => i.set_painted(),
         }
     }
 
-    pub fn on_displayed(&mut self, serial: FrameId) -> bool {
+    pub fn set_next_frame(&mut self) {
         match self {
-            Self::Strings(..) => false,
-            Self::Circle(i) => i.on_displayed(serial),
-            Self::Classic(..) => false,
+            Self::Strings(..) => {}
+            Self::Circle(i) => i.set_next_frame(),
+            Self::Classic(..) => {}
         }
     }
 
@@ -620,8 +619,11 @@ impl Button {
             self.width + 2.0,
             self.height + 2.0,
         );
+        cr.save().unwrap();
+        cr.set_operator(cairo::Operator::Source);
         cr.set_source(bg).unwrap();
         cr.fill().unwrap();
+        cr.restore().unwrap();
     }
 
     fn calc_extents(&mut self) {
@@ -802,6 +804,7 @@ pub fn setlocale() {
 #[derive(Debug)]
 pub struct Dialog {
     background: Pattern,
+    background_original: Rgba,
     buttons: Vec<Button>,
     labels: Vec<Label>,
     pub indicator: Indicator,
@@ -813,6 +816,8 @@ pub struct Dialog {
     debug: bool,
     pub uses_cursor: bool,
     button_pressed: bool,
+    transparency: bool,
+    dirty: bool,
 }
 
 impl Dialog {
@@ -977,24 +982,49 @@ impl Dialog {
             height,
             mouse_middle_pressed: false,
             background: config.background.into(),
+            background_original: config.background,
             input_timeout_duration: config.input_timeout.map(Duration::from_secs),
             input_timeout: None,
             debug,
             uses_cursor,
             button_pressed: false,
+            transparency: true,
+            dirty: false,
         })
     }
 
-    pub fn on_displayed(&mut self, serial: FrameId) -> bool {
-        self.indicator.on_displayed(serial)
+    pub fn set_transparency(&mut self, enable: bool) {
+        if self.transparency == enable {
+            debug!("set_transparency: status not changed");
+            return;
+        }
+        if self.background_original.alpha == u8::MAX {
+            debug!("set_transparency: original background not transparent");
+            return;
+        }
+        debug!("set_transparency: {}", enable);
+        self.dirty = true;
+        self.transparency = enable;
+        if enable {
+            self.background = self.background_original.into();
+        } else {
+            let mut background = self.background_original;
+            background.alpha = u8::MAX;
+            self.background = background.into();
+        }
     }
 
-    pub fn set_painted(&mut self, serial: FrameId) {
-        trace!("update serial {:?}", serial);
-        self.indicator.set_painted(serial);
+    pub fn set_next_frame(&mut self) {
+        self.indicator.set_next_frame()
+    }
+
+    pub fn set_painted(&mut self) {
+        trace!("set_painted");
+        self.indicator.set_painted();
         for b in &mut self.buttons {
             b.set_painted();
         }
+        self.dirty = false;
     }
 
     pub fn dirty(&self) -> bool {
@@ -1006,10 +1036,14 @@ impl Dialog {
                 return true;
             }
         }
-        false
+        self.dirty
     }
 
     pub fn repaint(&self, cr: &cairo::Context) {
+        if self.dirty {
+            return self.init(cr);
+        }
+
         self.indicator.repaint(cr, &self.background);
         for (i, b) in self.buttons.iter().enumerate() {
             if b.dirty {
@@ -1027,12 +1061,13 @@ impl Dialog {
         (size.0.round() as u16, size.1.round() as u16)
     }
 
-    pub fn init(&mut self, cr: &cairo::Context, serial: FrameId) {
+    pub fn init(&self, cr: &cairo::Context) {
         // TODO can I preserve antialiasing without clearing the image first?
+        cr.set_operator(cairo::Operator::Source);
         cr.set_source(&self.background).unwrap();
         cr.paint().unwrap();
+        cr.set_operator(cairo::Operator::Over);
         self.paint(cr);
-        self.set_painted(serial);
     }
 
     fn paint(&self, cr: &cairo::Context) {
@@ -1099,6 +1134,7 @@ impl Dialog {
     }
 
     pub fn resize(&mut self, cr: &cairo::Context, width: u16, height: u16, surface_cleared: bool) {
+        cr.set_operator(cairo::Operator::Source);
         cr.set_source(&self.background).unwrap();
 
         // TODO put to clear()
@@ -1117,6 +1153,7 @@ impl Dialog {
             );
             cr.fill().unwrap();
         }
+        cr.set_operator(cairo::Operator::Over);
 
         let mut m = cr.matrix();
 
