@@ -1,16 +1,19 @@
 use log::{debug, trace, warn};
+use tokio::io::unix::AsyncFd;
 use tokio::time::Instant;
 use x11rb::connection::Connection as _;
-use x11rb::protocol::xproto::{self, ConnectionExt as _};
+use x11rb::connection::RequestConnection as _;
+use x11rb::protocol::xfixes::{self, ConnectionExt as _};
+use x11rb::protocol::xproto::{self, ConnectionExt as _, CursorWrapper, WindowWrapper};
 use x11rb::protocol::Event as XEvent;
 use zeroize::Zeroize;
 
 use crate::backbuffer::Backbuffer;
 use crate::dialog::{Action, Dialog};
-use crate::errors::{Error, Result};
+use crate::errors::{Error, Result, Unsupported};
 use crate::keyboard::{Keyboard, Keycode};
 use crate::secret::Passphrase;
-use crate::{Connection, XId};
+use crate::Connection;
 
 enum State {
     Continue,
@@ -35,28 +38,57 @@ impl Drop for Keypress {
 }
 
 pub struct XContext<'a> {
-    pub conn: &'a Connection,
+    pub xfd: &'a AsyncFd<Connection>,
     pub backbuffer: Backbuffer<'a>,
-    pub(super) window: xproto::Window,
+    pub(super) window: WindowWrapper<'a, Connection>,
     pub keyboard: Keyboard<'a>,
     pub(super) atoms: crate::AtomCollection,
-    pub(super) colormap: XId,
-    pub(super) own_colormap: bool,
     pub(super) width: u16,
     pub(super) height: u16,
     pub(super) grab_keyboard: bool,
     pub(super) startup_time: Instant,
     pub(super) keyboard_grabbed: bool,
     pub(super) first_expose_received: bool,
-    pub(super) input_cursor: Option<XId>,
+    pub(super) input_cursor: Option<CursorWrapper<'a, Connection>>,
+    pub(super) compositor_atom: Option<xproto::Atom>,
 }
 
 impl<'a> XContext<'a> {
+    pub fn conn(&self) -> &Connection {
+        self.xfd.get_ref()
+    }
+
+    pub fn init(&self) -> Result<()> {
+        if let Some(compositor_atom) = self.compositor_atom {
+            self.conn()
+                .extension_information(xfixes::X11_EXTENSION_NAME)?
+                .ok_or_else(|| Unsupported("x11 xfixes extension required".into()))?;
+            let (major, minor) = xfixes::X11_XML_VERSION;
+            let version_cookie = self.conn().xfixes_query_version(major, minor)?;
+            if log::log_enabled!(log::Level::Debug) {
+                let version = version_cookie.reply()?;
+                debug!(
+                    "xfixes version {}.{}",
+                    version.major_version, version.minor_version
+                );
+            }
+
+            self.conn().xfixes_select_selection_input(
+                self.window.window(),
+                compositor_atom,
+                xfixes::SelectionEventMask::SET_SELECTION_OWNER
+                    | xfixes::SelectionEventMask::SELECTION_WINDOW_DESTROY
+                    | xfixes::SelectionEventMask::SELECTION_CLIENT_CLOSE,
+            )?;
+        }
+        Ok(())
+    }
+
     pub async fn run_events(&mut self, mut dialog: Dialog) -> Result<Option<Passphrase>> {
-        tokio::pin! { let xevents_ready = self.conn.xfd.readable(); }
+        tokio::pin! { let xevents_ready = self.xfd.readable(); }
         dialog.init_events();
         loop {
-            self.conn.flush()?;
+            self.conn().flush()?;
             tokio::select! {
                 action = dialog.handle_events() => {
                     if matches!(action, Action::Cancel) {
@@ -64,7 +96,7 @@ impl<'a> XContext<'a> {
                     }
                 }
                 xevents_guard = &mut xevents_ready => {
-                    if let Some(xevent) = self.conn.poll_for_event()? {
+                    if let Some(xevent) = self.conn().poll_for_event()? {
                         //silly!("xevent {:?}", xevent);
                         match self.handle_xevent(&mut dialog, xevent)? {
                             State::Continue => {},
@@ -74,19 +106,17 @@ impl<'a> XContext<'a> {
                     } else {
                         xevents_guard.unwrap().clear_ready();
                     }
-                    xevents_ready.set(self.conn.xfd.readable());
+                    xevents_ready.set(self.xfd.readable());
                 }
             }
-            if dialog.dirty() {
-                self.backbuffer.update(&mut dialog)?;
-            }
+            self.backbuffer.commit(&mut dialog)?;
             tokio::task::yield_now().await;
         }
     }
 
     pub fn set_default_cursor(&self) -> Result<()> {
-        self.conn.change_window_attributes(
-            self.window,
+        self.conn().change_window_attributes(
+            self.window.window(),
             &xproto::ChangeWindowAttributesAux::new().cursor(x11rb::NONE),
         )?;
         Ok(())
@@ -94,10 +124,10 @@ impl<'a> XContext<'a> {
 
     pub fn set_input_cursor(&self) -> Result<()> {
         trace!("set input cursor");
-        if let Some(cursor) = self.input_cursor {
-            self.conn.change_window_attributes(
-                self.window,
-                &xproto::ChangeWindowAttributesAux::new().cursor(cursor),
+        if let Some(ref cursor) = self.input_cursor {
+            self.conn().change_window_attributes(
+                self.window.window(),
+                &xproto::ChangeWindowAttributesAux::new().cursor(cursor.cursor()),
             )?;
             trace!("input cursor set");
         }
@@ -106,8 +136,8 @@ impl<'a> XContext<'a> {
 
     pub fn paste_primary(&self) -> Result<()> {
         trace!("PRIMARY selection");
-        self.conn.convert_selection(
-            self.window,
+        self.conn().convert_selection(
+            self.window.window(),
             xproto::AtomEnum::PRIMARY.into(),
             self.atoms.UTF8_STRING,
             self.atoms.XSEL_DATA,
@@ -117,8 +147,8 @@ impl<'a> XContext<'a> {
     }
 
     pub fn paste_clipboard(&self) -> Result<()> {
-        self.conn.convert_selection(
-            self.window,
+        self.conn().convert_selection(
+            self.window.window(),
             self.atoms.CLIPBOARD,
             self.atoms.UTF8_STRING,
             self.atoms.XSEL_DATA,
@@ -137,7 +167,7 @@ impl<'a> XContext<'a> {
                     return Ok(State::Continue);
                 }
 
-                self.backbuffer.present()?;
+                self.backbuffer.exposed = true;
 
                 if !self.first_expose_received {
                     debug!(
@@ -148,10 +178,10 @@ impl<'a> XContext<'a> {
 
                     if self.grab_keyboard {
                         let grabbed = self
-                            .conn
+                            .conn()
                             .grab_keyboard(
                                 false,
-                                self.window,
+                                self.window.window(),
                                 x11rb::CURRENT_TIME,
                                 xproto::GrabMode::ASYNC,
                                 xproto::GrabMode::ASYNC,
@@ -174,7 +204,6 @@ impl<'a> XContext<'a> {
                     self.width = ev.width;
                     self.height = ev.height;
                     self.backbuffer.resize_requested = Some((ev.width, ev.height));
-                    self.backbuffer.update(dialog)?;
                 }
             }
             // minimized
@@ -236,7 +265,7 @@ impl<'a> XContext<'a> {
                     warn!("selection failed?");
                 } else {
                     let reply = self
-                        .conn
+                        .conn()
                         .get_property(
                             false,
                             sn.requestor,
@@ -288,7 +317,7 @@ impl<'a> XContext<'a> {
                 self.backbuffer.on_idle_notify(&ev);
             }
             XEvent::PresentCompleteNotify(ev) => {
-                dialog.on_displayed(self.backbuffer.on_vsync_completed(ev));
+                self.backbuffer.on_vsync_completed(ev);
             }
             XEvent::XkbStateNotify(key) => {
                 self.keyboard.update_mask(&key);
@@ -303,6 +332,10 @@ impl<'a> XContext<'a> {
                 debug!("xkb map notify");
                 self.keyboard.reload_keymap();
             }
+            XEvent::XfixesSelectionNotify(sn) => {
+                debug!("selection notify: {:?}", sn);
+                dialog.set_transparency(sn.subtype == xfixes::SelectionEvent::SET_SELECTION_OWNER);
+            }
             event => {
                 debug!("unexpected event {:?}", event);
             }
@@ -314,25 +347,22 @@ impl<'a> XContext<'a> {
 impl<'a> Drop for XContext<'a> {
     fn drop(&mut self) {
         if self.keyboard_grabbed {
-            if let Err(err) = self.conn.ungrab_keyboard(x11rb::CURRENT_TIME) {
+            if let Err(err) = self.conn().ungrab_keyboard(x11rb::CURRENT_TIME) {
                 debug!("ungrab keyboard failed: {}", err);
             }
         }
+        if let Some(compositor_atom) = self.compositor_atom {
+            if let Err(err) = xfixes::select_selection_input(
+                self.conn(),
+                self.window.window(),
+                compositor_atom,
+                0u32,
+            ) {
+                debug!("clear select selection failed: {}", err);
+            }
+        }
         debug!("dropping XContext");
-        if let Err(err) = self.conn.destroy_window(self.window) {
-            debug!("destroy window failed: {}", err);
-        }
-        if self.own_colormap {
-            if let Err(err) = self.conn.free_colormap(self.colormap) {
-                debug!("free colormap failed: {}", err);
-            }
-        }
-        if let Some(cursor) = self.input_cursor {
-            if let Err(err) = self.conn.free_cursor(cursor) {
-                debug!("free cursor failed: {}", err);
-            }
-        }
-        if let Err(err) = self.conn.flush() {
+        if let Err(err) = self.conn().flush() {
             debug!("conn flush failed: {}", err);
         }
     }
