@@ -1,5 +1,4 @@
 use std::convert::TryInto as _;
-use std::ops::Deref;
 use std::os::unix::ffi::OsStrExt as _;
 use std::path::{Path, PathBuf};
 
@@ -9,9 +8,12 @@ use tokio::io::unix::AsyncFd;
 use tokio::signal::unix::{signal, SignalKind};
 use tokio::time::Instant;
 use x11rb::connection::{Connection as _, RequestConnection as _};
-use x11rb::protocol::xproto::{self, ConnectionExt as _};
+use x11rb::protocol::xproto::{
+    self, ColormapWrapper, ConnectionExt as _, CursorWrapper, WindowWrapper,
+};
 use x11rb::{atom_manager, properties};
 // for change_propertyN()
+use x11rb::protocol::render::{self, ConnectionExt as _, PictType};
 use x11rb::wrapper::ConnectionExt as _;
 use x11rb::xcb_ffi::XCBConnection;
 
@@ -23,7 +25,7 @@ mod event;
 mod keyboard;
 mod secret;
 
-use errors::{anyhow, Context as _, Result};
+use errors::{Context as _, Result};
 use secret::Passphrase;
 
 pub const CLASS: &str = "SshAskpass";
@@ -52,80 +54,101 @@ atom_manager! {
 
 pub type XId = u32;
 
-#[derive(Debug)]
-pub struct Connection {
-    pub xfd: AsyncFd<XCBConnection>,
+pub type Connection = XCBConnection;
+
+/// Modified from https://github.com/psychon/x11rb/blob/master/cairo-example/src/main.rs
+/// Choose a visual to use. This function tries to find a depth=32 visual and falls back to the
+/// screen's default visual.
+fn choose_visual(conn: &Connection, screen_num: usize) -> Result<(u8, xproto::Visualid)> {
+    let depth = 32;
+    let screen = &conn.setup().roots[screen_num];
+
+    // Try to use XRender to find a visual with alpha support
+    let has_render = conn
+        .extension_information(render::X11_EXTENSION_NAME)?
+        .is_some();
+    if has_render {
+        let formats = conn.render_query_pict_formats()?.reply()?;
+        // Find the ARGB32 format that must be supported.
+        let format = formats
+            .formats
+            .iter()
+            .filter(|info| (info.type_, info.depth) == (PictType::DIRECT, depth))
+            .filter(|info| {
+                let d = info.direct;
+                (d.red_mask, d.green_mask, d.blue_mask, d.alpha_mask) == (0xff, 0xff, 0xff, 0xff)
+            })
+            .find(|info| {
+                let d = info.direct;
+                (d.red_shift, d.green_shift, d.blue_shift, d.alpha_shift) == (16, 8, 0, 24)
+            });
+        if let Some(format) = format {
+            // Now we need to find the visual that corresponds to this format
+            if let Some(visual) = formats.screens[screen_num]
+                .depths
+                .iter()
+                .flat_map(|d| &d.visuals)
+                .find(|v| v.format == format.id)
+            {
+                return Ok((format.depth, visual.visual));
+            }
+        }
+    }
+    Ok((screen.root_depth, screen.root_visual))
 }
 
-impl Connection {
-    pub fn new() -> Result<(Self, usize)> {
-        let (conn, screen_num) = XCBConnection::connect(None).context("X11 connect")?;
-        debug!("preferred screen {}", screen_num);
-        let me = Self {
-            // There are no reasonable failures, so lets panic
-            xfd: AsyncFd::new(conn).expect("asyncfd failed"),
-        };
-        Ok((me, screen_num))
-    }
-}
-
-impl Deref for Connection {
-    type Target = XCBConnection;
-    fn deref(&self) -> &Self::Target {
-        self.xfd.get_ref()
+/// Find a `xcb_visualtype_t` based on its ID number
+fn find_xcb_visualtype(conn: &Connection, visual_id: u32) -> Option<xproto::Visualtype> {
+    for root in &conn.setup().roots {
+        for depth in &root.allowed_depths {
+            for visual in &depth.visuals {
+                if visual.visual_id == visual_id {
+                    return Some(*visual);
+                }
+            }
+        }
     }
+    None
 }
 
 async fn run_xcontext(
-    cfg_loader: config::Loader,
+    config: config::Config,
     opts: Opts,
     startup_time: Instant,
 ) -> Result<Option<Passphrase>> {
-    let (conn, screen_num) = Connection::new()?;
+    let (conn, screen_num) = XCBConnection::connect(None).context("X11 connect")?;
+    let xfd = AsyncFd::new(conn).context("asyncfd failed")?;
+    let conn = xfd.get_ref();
+
     debug!("connected X server");
-    let atoms = AtomCollection::new(&*conn)?;
-    debug!("loaded atoms");
+    let atoms = AtomCollection::new(conn)?;
+    let compositor_atom = if config.depth == 32 {
+        conn.prefetch_extension_information(x11rb::protocol::xfixes::X11_EXTENSION_NAME)?;
+        let compositor_atom = format!("_NET_WM_CM_S{}", screen_num);
+        Some(conn.intern_atom(false, compositor_atom.as_bytes())?)
+    } else {
+        None
+    };
 
     conn.prefetch_extension_information(x11rb::protocol::present::X11_EXTENSION_NAME)?;
+    conn.prefetch_extension_information(x11rb::protocol::xkb::X11_EXTENSION_NAME)?;
+    conn.prefetch_extension_information(x11rb::protocol::render::X11_EXTENSION_NAME)?;
 
     conn.flush()?;
 
     let setup = conn.setup();
     let screen = setup.roots.get(screen_num).expect("unknown screen");
 
-    let instance = opts.name.unwrap_or_else(|| {
-        std::env::args()
-            .next()
-            .as_ref()
-            .and_then(|f| Path::new(f).file_name()?.to_str())
-            .unwrap()
-            .to_string()
-    });
-
-    debug!("load config");
-    let config = if let Some(path) = opts.config {
-        cfg_loader.load_path(&path)?
+    let (depth, visualid) = if config.depth == 32 {
+        choose_visual(conn, screen_num)?
     } else {
-        cfg_loader.load()?
+        (screen.root_depth, screen.root_visual)
     };
-    debug!("config loaded");
-
-    // TODO where are the expose events with depth 32?
-    let depth = config.depth;
     debug!("window depth: {}", depth);
+    let visual_type = find_xcb_visualtype(conn, visualid).unwrap();
 
-    let depth_type = screen
-        .allowed_depths
-        .iter()
-        .find(|d| d.depth == depth)
-        .ok_or_else(|| anyhow!("invalid depth"))?;
-    let visual_type = depth_type
-        .visuals
-        .get(0)
-        .ok_or_else(|| anyhow!("depth has no visual types"))?;
-
-    let surface = backbuffer::XcbSurface::new(&conn, screen.root, depth, visual_type, 1, 1)?;
-    let backbuffer = backbuffer::Backbuffer::new(&conn, screen.root, surface)?;
+    let surface = backbuffer::XcbSurface::new(conn, screen.root, depth, &visual_type, 1, 1)?;
+    let backbuffer = backbuffer::Backbuffer::new(conn, screen.root, surface)?;
     conn.flush()?;
     let mut dialog = dialog::Dialog::new(
         config.dialog,
@@ -139,23 +162,21 @@ async fn run_xcontext(
     debug!("window width: {}, height: {}", window_width, window_height);
 
     let colormap = if visual_type.visual_id == screen.root_visual {
-        screen.default_colormap
+        None
     } else {
         debug!("depth requires a new colormap");
-        let colormap = conn.generate_id()?;
-        conn.create_colormap(
+        let colormap = ColormapWrapper::create_colormap(
+            conn,
             xproto::ColormapAlloc::NONE,
-            colormap,
             screen.root,
             visual_type.visual_id,
         )?;
-        colormap
+        Some(colormap)
     };
 
-    let window = conn.generate_id()?;
-    conn.create_window(
+    let window_wrapper = WindowWrapper::create_window(
+        conn,
         depth,
-        window,
         screen.root,
         0, // x
         0, // y
@@ -176,8 +197,14 @@ async fn run_xcontext(
             )
             .background_pixmap(xproto::PixmapEnum::NONE)
             .border_pixel(screen.black_pixel)
-            .colormap(colormap),
+            .colormap(
+                colormap
+                    .as_ref()
+                    .map(|c| c.colormap())
+                    .unwrap_or(screen.default_colormap),
+            ),
     )?;
+    let window = window_wrapper.window();
 
     let atoms = atoms.reply()?;
 
@@ -187,8 +214,10 @@ async fn run_xcontext(
         gethostname::gethostname()
     };
     let mut title = config.title;
-    title.push('@');
-    title.push_str(&hostname.to_string_lossy());
+    if config.show_hostname {
+        title.push('@');
+        title.push_str(&hostname.to_string_lossy());
+    }
     conn.change_property8(
         xproto::PropMode::REPLACE,
         window,
@@ -222,7 +251,7 @@ async fn run_xcontext(
         window,
         xproto::AtomEnum::WM_CLASS,
         xproto::AtomEnum::STRING,
-        [instance.as_bytes(), CLASS.as_bytes()]
+        [opts.instance().as_bytes(), CLASS.as_bytes()]
             .join(&b'\0')
             .as_slice(),
     )?;
@@ -271,7 +300,7 @@ async fn run_xcontext(
         ..properties::WmHints::default()
     };
     // TODO icon?
-    wm_hints.set(&*conn, window)?;
+    wm_hints.set(conn, window)?;
 
     for (width, height, data) in ICONS {
         let mut icon_data = Vec::with_capacity(8 + data.len());
@@ -301,7 +330,7 @@ async fn run_xcontext(
     if !config.resizable {
         size_hints.max_size = Some((window_width.into(), window_height.into()));
     }
-    size_hints.set_normal_hints(&*conn, window)?;
+    size_hints.set_normal_hints(conn, window)?;
 
     debug!("map window");
     conn.map_window(window)?;
@@ -310,42 +339,51 @@ async fn run_xcontext(
 
     // Load the slow ones after we have mapped the window
 
+    let (transparency, compositor_atom) = if let Some(compositor_atom) = compositor_atom {
+        let compositor_atom = compositor_atom.reply()?.atom;
+        let selection = conn.get_selection_owner(compositor_atom)?;
+        (
+            selection.reply()?.owner != x11rb::NONE,
+            Some(compositor_atom),
+        )
+    } else {
+        (false, None)
+    };
+
     let resource_db;
     let cursor_handle = if dialog.uses_cursor {
-        resource_db = x11rb::resource_manager::Database::new_from_default(&*conn)?;
-        Some(x11rb::cursor::Handle::new(
-            &*conn,
-            screen_num,
-            &resource_db,
-        )?)
+        resource_db = x11rb::resource_manager::Database::new_from_default(conn)?;
+        Some(x11rb::cursor::Handle::new(conn, screen_num, &resource_db)?)
     } else {
         None
     };
 
+    debug!("compositor detected: {}", transparency);
+    dialog.set_transparency(transparency);
+
     debug!("dialog init");
     let mut backbuffer = backbuffer.reply()?;
     backbuffer.init(window, &mut dialog)?;
 
     debug!("keyboard init");
-    let keyboard = keyboard::Keyboard::new(&conn)?;
+    let keyboard = keyboard::Keyboard::new(conn)?;
 
     debug!("cursor init");
     let input_cursor = if let Some(cursor_handle) = cursor_handle {
         let cursor_handle = cursor_handle.reply()?;
-        Some(cursor_handle.load_cursor(&*conn, "xterm").unwrap())
+        Some(CursorWrapper::for_cursor(
+            conn,
+            cursor_handle.load_cursor(conn, "xterm").unwrap(),
+        ))
     } else {
         None
     };
-    debug!("init took {}ms", startup_time.elapsed().as_millis());
-
     let mut xcontext = event::XContext {
         keyboard,
-        conn: &conn,
+        xfd: &xfd,
         backbuffer,
-        window,
+        window: window_wrapper,
         atoms,
-        colormap,
-        own_colormap: colormap != screen.default_colormap,
         width: window_width,
         height: window_height,
         grab_keyboard: config.grab_keyboard,
@@ -353,8 +391,12 @@ async fn run_xcontext(
         first_expose_received: false,
         keyboard_grabbed: false,
         input_cursor,
+        compositor_atom,
     };
 
+    xcontext.init()?;
+    debug!("init took {}ms", startup_time.elapsed().as_millis());
+
     xcontext.run_events(dialog).await
 }
 
@@ -394,6 +436,19 @@ struct Opts {
     gen_config: bool,
 }
 
+impl Opts {
+    fn instance(&self) -> String {
+        self.name.clone().unwrap_or_else(|| {
+            std::env::args()
+                .next()
+                .as_ref()
+                .and_then(|f| Path::new(f).file_name()?.to_str())
+                .unwrap()
+                .to_string()
+        })
+    }
+}
+
 fn run() -> i32 {
     let startup_time = Instant::now();
 
@@ -408,7 +463,9 @@ fn run() -> i32 {
     let opts = Opts::from_arg_matches(&app.get_matches()).expect("from_arg_matches");
 
     let mut log = stderrlog::new();
-    log.quiet(opts.quiet).verbosity(opts.verbose + 2);
+    log.quiet(opts.quiet)
+        .verbosity(opts.verbose + 2)
+        .show_module_names(true);
     if opts.debug {
         log.timestamp(stderrlog::Timestamp::Millisecond);
     }
@@ -430,6 +487,14 @@ fn run_logged(cfg_loader: config::Loader, opts: Opts, startup_time: Instant) ->
         return Ok(0);
     }
 
+    debug!("load config");
+    let config = if let Some(ref path) = opts.config {
+        cfg_loader.load_path(path)?
+    } else {
+        cfg_loader.load()?
+    };
+    debug!("config loaded");
+
     dialog::setlocale();
 
     let runtime = tokio::runtime::Builder::new_current_thread()
@@ -457,7 +522,7 @@ fn run_logged(cfg_loader: config::Loader, opts: Opts, startup_time: Instant) ->
             _ = sigterm.recv() => {
                 info!("got sigterm");
             }
-            ret = run_xcontext(cfg_loader, opts, startup_time) => {
+            ret = run_xcontext(config, opts, startup_time) => {
                 match ret? {
                     Some(pass) => {
                         pass.write_stdout().unwrap();
